"""
Definition of Fluid, IncompressibleFlow as well as fluid-related functions.
"""
from numbers import Number

import numpy as np

from phi import struct
from phi import math

from .physics import StateDependency, Physics
from .pressuresolver.solver_api import FluidDomain
from .pressuresolver.sparse import SparseCG
from .field import CenteredGrid, StaggeredGrid, union_mask, advect
from .material import OPEN, Material
from .domain import Domain, DomainState
from .field.effect import Gravity, gravity_tensor, effect_applied


@struct.definition()
class Fluid(DomainState):
    """
    A Fluid state consists of a density field (centered grid) and a velocity field (staggered grid).
    """

    def __init__(self, domain, density=0.0, velocity=0.0, buoyancy_factor=0.0, tags=('fluid', 'velocityfield'), name='fluid', **kwargs):
        DomainState.__init__(self, **struct.kwargs(locals()))

    def default_physics(self): return INCOMPRESSIBLE_FLOW

    @struct.variable(default=0, dependencies=DomainState.domain)
    def density(self, density):
        """
The marker density is stored in a CenteredGrid with dimensions matching the domain.
It describes the number of particles per physical volume.
        """
        return self.centered_grid('density', density)

    @struct.variable(default=0, dependencies=DomainState.domain)
    def velocity(self, velocity):
        """
The velocity is stored in a StaggeredGrid with dimensions matching the domain.
        """
        return self.staggered_grid('velocity', velocity)

    @struct.constant(default=0.0)
    def buoyancy_factor(self, fac):
        """
The default fluid physics can apply Boussinesq buoyancy as an upward force, proportional to the density.
This force is scaled with the buoyancy_factor (float).
        """
        return fac

    def __repr__(self):
        return "Fluid[density: %s, velocity: %s]" % (self.density, self.velocity)


class IncompressibleFlow(Physics):
    """
Physics modelling the incompressible Navier-Stokes equations.
Supports buoyancy proportional to the marker density.
Supports obstacles, density effects, velocity effects, global gravity.
    """

    def __init__(self, pressure_solver=None, make_input_divfree=False, make_output_divfree=True, conserve_density=True):
        Physics.__init__(self, [StateDependency('obstacles', 'obstacle'),
                                StateDependency('gravity', 'gravity', single_state=True),
                                StateDependency('density_effects', 'density_effect', blocking=True),
                                StateDependency('velocity_effects', 'velocity_effect', blocking=True)])
        self.pressure_solver = pressure_solver
        self.make_input_divfree = make_input_divfree
        self.make_output_divfree = make_output_divfree
        self.conserve_density = conserve_density

    def step(self, fluid, dt=1.0, obstacles=(), gravity=Gravity(), density_effects=(), velocity_effects=()):
        # pylint: disable-msg = arguments-differ
        gravity = gravity_tensor(gravity, fluid.rank)
        velocity = fluid.velocity
        density = fluid.density
        if self.make_input_divfree:
            velocity = divergence_free(velocity, fluid.domain, obstacles, pressure_solver=self.pressure_solver)
        # --- Advection ---
        density = advect.semi_lagrangian(density, velocity, dt=dt)
        velocity = advect.semi_lagrangian(velocity, velocity, dt=dt)
        if self.conserve_density and np.all(Material.solid(fluid.domain.boundaries)):
            density = density.normalized(fluid.density)
        # --- Effects ---
        for effect in density_effects:
            density = effect_applied(effect, density, dt)
        for effect in velocity_effects:
            velocity = effect_applied(effect, velocity, dt)
        velocity += buoyancy(fluid.density, gravity, fluid.buoyancy_factor) * dt
        # --- Pressure solve ---
        if self.make_output_divfree:
            velocity = divergence_free(velocity, fluid.domain, obstacles, pressure_solver=self.pressure_solver)
        return fluid.copied_with(density=density, velocity=velocity, age=fluid.age + dt)


INCOMPRESSIBLE_FLOW = IncompressibleFlow()


def buoyancy(density, gravity, buoyancy_factor):
    """
Computes the buoyancy force proportional to the density.
    :param density: CenteredGrid
    :param gravity: vector or float
    :param buoyancy_factor: float
    :return: StaggeredGrid for the domain of the density
    """
    if isinstance(gravity, (int, float)):
        gravity = math.to_float(math.as_tensor([gravity] + ([0] * (density.rank - 1))))
    result = StaggeredGrid.from_scalar(density, -gravity * buoyancy_factor)
    return result


def _is_div_free(velocity, is_div_free):
    assert is_div_free in (True, False, None)
    if isinstance(is_div_free, bool):
        return is_div_free
    if isinstance(velocity, Number):
        return True
    return False


def solve_pressure(divergence, fluiddomain, pressure_solver=None):
    """
    Computes the pressure from the given velocity or velocity divergence using the specified solver.
    :param divergence: CenteredGrid
    :param fluiddomain: FluidDomain instance
    :param pressure_solver: PressureSolver to use, None for default
    :return: scalar tensor or CenteredGrid, depending on the type of divergence
    """
    assert isinstance(divergence, CenteredGrid)
    if pressure_solver is None:
        pressure_solver = SparseCG()
    pressure, iteration = pressure_solver.solve(divergence.data, fluiddomain, pressure_guess=None)
    if isinstance(divergence, CenteredGrid):
        pressure = CenteredGrid(pressure, divergence.box, name='pressure')
    return pressure, iteration


def divergence_free(velocity, domain=None, obstacles=(), pressure_solver=None):
    """
Projects the given velocity field by solving for and subtracting the pressure.
    :param velocity: StaggeredGrid
    :param domain: Domain matching the velocity field, used for boundary conditions
    :param obstacles: list of Obstacles
    :param pressure_solver: PressureSolver. Uses default solver if none provided.
    :return: divergence-free velocity as StaggeredGrid
    """
    assert isinstance(velocity, StaggeredGrid)
    # --- Set up FluidDomain ---
    if domain is None:
        domain = Domain(velocity.resolution, OPEN)
    obstacle_mask = union_mask([obstacle.geometry for obstacle in obstacles])
    if obstacle_mask is not None:
        obstacle_grid = obstacle_mask.at(velocity.center_points, collapse_dimensions=False).copied_with(extrapolation='constant')
        active_mask = 1 - obstacle_grid
    else:
        active_mask = math.ones(domain.centered_shape(name='active', extrapolation='constant'))
    accessible_mask = active_mask.copied_with(extrapolation=Material.accessible_extrapolation_mode(domain.boundaries))
    fluiddomain = FluidDomain(domain, active=active_mask, accessible=accessible_mask)
    # --- Boundary Conditions, Pressure Solve ---
    velocity = fluiddomain.with_hard_boundary_conditions(velocity)
    divergence_field = velocity.divergence(physical_units=False)
    pressure, _ = solve_pressure(divergence_field, fluiddomain, pressure_solver=pressure_solver)
    pressure *= velocity.dx[0]
    gradp = StaggeredGrid.gradient(pressure)
    velocity -= fluiddomain.with_hard_boundary_conditions(gradp)
    return velocity



from .field import Field
from phi import struct, math


@struct.definition()
class AnalyticField(Field):

    def __init__(self, rank, data=1.0, name=None, **kwargs):
        Field.__init__(self, **struct.kwargs(locals(), ignore='rank'))
        self._rank = rank

    @property
    def rank(self):
        return self._rank

    @property
    def component_count(self):
        example_value = self.sample_at([[0] * self._rank])
        return math.staticshape(example_value)[-1]

    def unstack(self):
        if self.component_count == 1:
            return [self]
        else:
            raise NotImplementedError()

    @property
    def points(self):
        return None

    def compatible(self, other_field):
        return True



import logging
import os
import numpy as np
import tensorflow as tf
from numbers import Number

from phi import math
from phi.physics.pressuresolver.solver_api import PressureSolver


if tf.__version__[0] == '2':
    logging.info('Adjusting for tensorflow 2.0')
    tf = tf.compat.v1
    tf.disable_eager_execution()


# Load Custom Ops
current_dir = os.path.dirname(os.path.realpath(__file__))
kernel_path = os.path.join(current_dir, 'cuda/build/pressure_solve_op.so')
assert os.path.isfile(kernel_path), 'CUDA binaries not found at %s. Run "python setup.py cuda" to compile them' % kernel_path
pressure_op = tf.load_op_library(kernel_path)


class CUDASolver(PressureSolver):

    def __init__(self, accuracy=1e-5, gradient_accuracy='same',
                 max_iterations=2000, max_gradient_iterations='same'):
        PressureSolver.__init__(self, 'CUDA Conjugate Gradient', supported_devices=('GPU',),
                                supports_loop_counter=True, supports_guess=False, supports_continuous_masks=False)
        assert isinstance(accuracy, Number), 'invalid accuracy: %s' % accuracy
        assert gradient_accuracy == 'same' or isinstance(gradient_accuracy, Number), 'invalid gradient_accuracy: %s' % gradient_accuracy
        assert max_gradient_iterations in ['same', 'mirror'] or isinstance(max_gradient_iterations, Number), 'invalid max_gradient_iterations: %s' % max_gradient_iterations
        self.accuracy = accuracy
        self.gradient_accuracy = accuracy if gradient_accuracy == 'same' else gradient_accuracy
        self.max_iterations = max_iterations
        if max_gradient_iterations == 'same':
            self.max_gradient_iterations = max_iterations
        elif max_gradient_iterations == 'mirror':
            self.max_gradient_iterations = 'mirror'
        else:
            self.max_gradient_iterations = max_gradient_iterations

    def solve(self, divergence, domain, pressure_guess):
        # pressure_guess: not used in this implementation, Kernel takes the last pressure value for initial_guess
        active, accessible = domain.active_tensor(extend=1), domain.accessible_tensor(extend=1)

        def pressure_gradient(op, grad):
            return cuda_solve_forward(grad, active, accessible, self.gradient_accuracy, max_gradient_iterations)[0]

        pressure, iteration = math.with_custom_gradient(
            cuda_solve_forward,
            [divergence, active, accessible, self.accuracy, self.max_iterations],
            pressure_gradient,
            input_index=0, output_index=0, name_base='cuda_pressure_solve'
        )
        max_gradient_iterations = iteration if self.max_gradient_iterations == 'mirror' else self.max_gradient_iterations
        return pressure, iteration


def cuda_solve_forward(divergence, active_mask, fluid_mask, accuracy, max_iterations):
    # Setup
    dimensions = divergence.get_shape()[1:-1]
    dimensions = dimensions[::-1]  # the custom op needs it in the x,y,z order
    dim_array = np.array(dimensions)
    dim_product = np.prod(dimensions)
    mask_dimensions = dim_array + 2
    laplace_matrix = tf.zeros(dim_product * (len(dimensions) * 2 + 1), dtype=tf.int8)
    # Helper variables for CG, make sure new memory is allocated for each variable.
    one_vector = tf.ones(dim_product, dtype=tf.float32)
    p = tf.zeros_like(divergence, dtype=tf.float32) + 1
    z = tf.zeros_like(divergence, dtype=tf.float32) + 2
    r = tf.zeros_like(divergence, dtype=tf.float32) + 3
    pressure = tf.zeros_like(divergence, dtype=tf.float32) + 4
    # Solve
    pressure, iteration = pressure_op.pressure_solve(
        dimensions, mask_dimensions, active_mask, fluid_mask, laplace_matrix,
        divergence, p, r, z, pressure, one_vector, dim_product, accuracy, max_iterations
    )
    return pressure, iteration



from .source import DataSource, SceneSource
from weakref import WeakSet


class Dataset(object):

    def __init__(self, name):
        self.name = name
        self.sources = []
        self._observers = WeakSet()

    def __repr__(self):
        return self.name

    def add(self, datasource):
        if isinstance(datasource, DataSource):
            self.sources.append(datasource)
        else:
            for source in datasource:
                self.add(source)
        for o in self._observers:
            o(self)

    def remove(self, datasource):
        self.sources.remove(datasource)
        for o in self._observers:
            o(self)

    def count(self, lookup_unknown=True):
        total = 0
        for datasource in self.sources:
            s = datasource.size(lookup=lookup_unknown)
            if s is not None:
                total += s
        return total

    def on_change(self, callback):
        self._observers.add(callback)

    def remove_on_change(self, callback):
        self._observers.remove(callback)

    def __iadd__(self, other):
        if isinstance(other, DataSource):
            self.add(other)
        if isinstance(other, Dataset):
            self.sources = self.sources + other.sources
            for o in self._observers:
                o(self)
        return self

    def __add__(self, other):
        newset = Dataset("%s + %s" % (self.name, other.name))
        if isinstance(other, DataSource):
            newset.add(other)
        if isinstance(other, Dataset):
            newset.sources = self.sources + other.sources
        return newset

    @staticmethod
    def load(directory, indices=None, name=None, max_scenes=None, assume_same_frames=True, assume_same_shapes=True):
        import os
        from .fluidformat import Scene

        if name is None:
            name = os.path.basename(directory)

        dataset = Dataset(name)

        shape_map = dict() if assume_same_shapes else None
        frames = None

        indexfilter = None if indices is None else lambda i: i in indices
        scene_iterator = Scene.list(directory, max_count=max_scenes, indexfilter=indexfilter)

        for scene in scene_iterator:
            if assume_same_frames and frames is None:
                frames = scene.frames
            dataset.add(SceneSource(scene, frames=frames, shape_map=shape_map))

        if dataset.count()==0:
            raise ValueError("No data sets found in '%s' " % directory)

        return dataset



# coding=utf-8
import logging
import warnings
import numpy as np
import tensorflow as tf
from tensorflow.python import pywrap_tensorflow

from phi import struct
from phi.math.math_util import is_static_shape
from phi.physics.field.staggered_grid import StaggeredGrid
from phi.physics.field.grid import CenteredGrid
from phi.struct.functions import mappable

if tf.__version__[0] == '2':
    logging.info('Adjusting for tensorflow 2.0')
    tf = tf.compat.v1
    tf.disable_eager_execution()


def _tf_name(trace, basename):
    if basename is None:
        return trace.path('/')
    else:
        return basename + '/' + trace.path('/')


def placeholder(shape, dtype=np.float32, basename=None, item_condition=struct.VARIABLES):
    if struct.isstruct(dtype):
        def placeholder_map(trace):
            shape, dtype = trace.value
            return tf.placeholder(dtype, shape, _tf_name(trace, basename))
        zipped = struct.zip([shape, dtype], leaf_condition=is_static_shape, item_condition=item_condition)
        return struct.map(placeholder_map, zipped, leaf_condition=is_static_shape, trace=True, item_condition=item_condition)
    else:
        f = lambda trace: tf.placeholder(dtype, trace.value, _tf_name(trace, basename))
        return struct.map(f, shape, leaf_condition=is_static_shape, trace=True, item_condition=item_condition)


def placeholder_like(obj, basename=None):
    warnings.warn("placeholder_like may not respect the batch dimension. "
                  "For State objects, use placeholder(state.shape) instead.", DeprecationWarning, stacklevel=2)
    f = lambda attr: tf.placeholder(attr.value.dtype, attr.value.shape, _tf_name(attr, basename))
    return struct.map(f, obj, leaf_condition=is_static_shape, trace=True)


def variable(initial_value, dtype=np.float32, basename=None, trainable=True, item_condition=struct.VARIABLES):
    f = lambda attr: tf.Variable(attr.value, name=_tf_name(attr, basename), dtype=dtype, trainable=trainable)
    return struct.map(f, initial_value, leaf_condition=is_static_shape, trace=True, item_condition=item_condition)


def variable_generator(initializer, dtype=np.float32, basename=None, trainable=True):
    def create_variable(shape):
        initial_value = initializer(shape)
        return variable(initial_value, dtype, basename, trainable)
    return create_variable


def isplaceholder(obj):
    return isinstance(obj, tf.Tensor) and obj.op.type == 'Placeholder'


def group_normalization(x, group_count, eps=1e-5):
    batch_size, H, W, C = tf.shape(x)
    gamma = tf.Variable(np.ones([1, 1, 1, C]), dtype=tf.float32, name="GN_gamma")
    beta = tf.Variable(np.zeros([1, 1, 1, C]), dtype=tf.float32, name="GN_beta")
    x = tf.reshape(x, [batch_size, group_count, H, W, C // group_count])
    mean, var = tf.nn.moments(x, [2, 3, 4], keep_dims=True)
    x = (x - mean) / tf.sqrt(var + eps)
    x = tf.reshape(x, [batch_size, H, W, C])
    return x * gamma + beta


def residual_block(y, nb_channels, kernel_size=(3, 3), _strides=(1, 1), activation=tf.nn.leaky_relu,
                   _project_shortcut=False, padding="SYMMETRIC", name=None, training=False, trainable=True, reuse=tf.AUTO_REUSE):
    shortcut = y

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)

    pad1 = [(kernel_size[0] - 1) // 2, kernel_size[0] // 2]
    pad2 = [(kernel_size[1] - 1) // 2, kernel_size[1] // 2]

    # down-sampling is performed with a stride of 2
    y = tf.pad(y, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
    y = tf.layers.conv2d(y, nb_channels, kernel_size=kernel_size, strides=_strides, padding='valid',
                         name=None if name is None else name+"/conv1", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm1", training=training, trainable=trainable, reuse=reuse)
    y = activation(y)

    y = tf.pad(y, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
    y = tf.layers.conv2d(y, nb_channels, kernel_size=kernel_size, strides=(1, 1), padding='valid',
                         name=None if name is None else name + "/conv2", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm2", training=training, trainable=trainable, reuse=reuse)

    # identity shortcuts used directly when the input and output are of the same dimensions
    if _project_shortcut or _strides != (1, 1):
        # when the dimensions increase projection shortcut is used to match dimensions (done by 1×1 convolutions)
        # when the shortcuts go across feature maps of two sizes, they are performed with a stride of 2
        shortcut = tf.pad(shortcut, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
        shortcut = tf.layers.conv2d(shortcut, nb_channels, kernel_size=(1, 1), strides=_strides, padding='valid',
                                    name=None if name is None else name + "/convid", trainable=trainable, reuse=reuse)
        # shortcut = tf.layers.batch_normalization(shortcut, name=None if name is None else name+"/normid", training=training, trainable=trainable, reuse=reuse)

    y += shortcut
    y = activation(y)

    return y


def residual_block_1d(y, nb_channels, kernel_size=(3,), _strides=(1,), activation=tf.nn.leaky_relu,
                   _project_shortcut=False, padding="SYMMETRIC", name=None, training=False, trainable=True, reuse=tf.AUTO_REUSE):
    shortcut = y

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size,)

    pad1 = [(kernel_size[0] - 1) // 2, kernel_size[0] // 2]

    # down-sampling is performed with a stride of 2
    y = tf.pad(y, [[0, 0], pad1, [0, 0]], mode=padding)
    y = tf.layers.conv1d(y, nb_channels, kernel_size=kernel_size, strides=_strides, padding='valid',
             name=None if name is None else name+"/conv1", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm1", training=training, trainable=trainable, reuse=reuse)
    y = activation(y)

    y = tf.pad(y, [[0, 0], pad1, [0, 0]], mode=padding)
    y = tf.layers.conv1d(y, nb_channels, kernel_size=kernel_size, strides=(1,), padding='valid',
             name=None if name is None else name + "/conv2", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm2", training=training, trainable=trainable, reuse=reuse)

    # identity shortcuts used directly when the input and output are of the same dimensions
    if _project_shortcut or _strides != (1,):
        # when the dimensions increase projection shortcut is used to match dimensions (done by 1×1 convolutions)
        # when the shortcuts go across feature maps of two sizes, they are performed with a stride of 2
        shortcut = tf.pad(shortcut, [[0, 0], pad1, [0, 0]], mode=padding)
        shortcut = tf.layers.conv1d(shortcut, nb_channels, kernel_size=(1, 1), strides=_strides, padding='valid',
                        name=None if name is None else name + "/convid", trainable=trainable, reuse=reuse)
        # shortcut = tf.layers.batch_normalization(shortcut, name=None if name is None else name+"/normid", training=training, trainable=trainable, reuse=reuse)

    y += shortcut
    y = activation(y)

    return y


def istensor(obj):
    if isinstance(obj, CenteredGrid):
        return istensor(obj.data)
    if isinstance(obj, StaggeredGrid):
        return np.any([istensor(t) for t in obj.data])
    return isinstance(obj, (tf.Tensor, tf.Variable))



def conv_function(scope, constants_file=None):
    if constants_file is not None:
        reader = pywrap_tensorflow.NewCheckpointReader(constants_file)
        def conv(n, filters, kernel_size, strides=[1, 1, 1, 1], padding="VALID", activation=None, name=None, kernel_initializer=None):
            assert name is not None
            kernel = reader.get_tensor("%s/%s/kernel" % (scope, name))
            assert kernel.shape[-1] == filters, "Expected %d filters but loaded kernel has shape %s for conv %s" % (kernel_size, kernel.shape, name)
            if isinstance(kernel_size, int):
                assert kernel.shape[0] == kernel.shape[1] == kernel_size
            else:
                assert kernel.shape[0:2] == kernel_size
            if isinstance(strides, int):
                strides = [1, strides, strides, 1]
            elif len(strides) == 2:
                strides = [1, strides[0], strides[1], 1]
            n = tf.nn.conv2d(n, kernel, strides=strides, padding=padding.upper(), name=name)
            if activation is not None:
                n = activation(n)
            n = tf.nn.bias_add(n, reader.get_tensor("%s/%s/bias" % (scope, name)))
            return n
    else:
        def conv(n, filters, kernel_size, strides=(1, 1), padding="valid", activation=None, name=None, kernel_initializer=None):
            with tf.variable_scope(scope):
                return tf.layers.conv2d(n, filters=filters, kernel_size=kernel_size, strides=strides, padding=padding,
                                        activation=activation, name=name, reuse=tf.AUTO_REUSE, kernel_initializer=kernel_initializer)
    return conv



import numpy as np

from phi import struct
from phi import math


@struct.definition()
class Geometry(struct.Struct):

    def value_at(self, location):
        """
Samples the geometry at the given locations and returns a binary mask, labelling the points as inside=1, outside=0.
        :param location: tensor of the shape (batch_size, ..., rank)
        :return: float tensor of same shape as location but with shape[-1]=1
        """
        raise NotImplementedError(self.__class__)

    @property
    def rank(self):
        raise NotImplementedError()


@struct.definition()
class AABox(Geometry):

    def __init__(self, lower, upper, **kwargs):
        Geometry.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def lower(self, lower):
        return math.to_float(math.as_tensor(lower))

    @struct.constant()
    def upper(self, upper):
        return math.to_float(math.as_tensor(upper))

    def get_lower(self, axis):
        return self._get(self.lower, axis)

    def get_upper(self, axis):
        return self._get(self.upper, axis)

    @staticmethod
    def _get(vector, axis):
        if math.ndims(vector) == 0:
            return vector
        else:
            return vector[...,axis]

    @struct.derived()
    def size(self):
        return self.upper - self.lower

    @property
    def rank(self):
        if math.ndims(self.size) > 0:
            return self.size.shape[-1]
        else:
            return None

    def global_to_local(self, global_position):
        size, lower = math.batch_align([self.size, self.lower], 1, global_position)
        return (global_position - lower) / size

    def local_to_global(self, local_position):
        size, lower = math.batch_align([self.size, self.lower], 1, local_position)
        return local_position * size + lower

    def value_at(self, global_position):
        lower, upper = math.batch_align([self.lower, self.upper], 1, global_position)
        bool_inside = (global_position >= lower) & (global_position <= upper)
        bool_inside = math.all(bool_inside, axis=-1, keepdims=True)
        return math.to_float(bool_inside)

    def contains(self, other):
        if isinstance(other, AABox):
            return np.all(other.lower >= self.lower) and np.all(other.upper <= self.upper)
        else:
            raise NotImplementedError()

    def without_axis(self, axis):
        lower = []
        upper = []
        for ax in range(self.rank):
            if ax != axis:
                lower.append(self.get_lower(ax))
                upper.append(self.get_upper(ax))
        return self.copied_with(lower=lower, upper=upper)

    def __repr__(self):
        try:
            return '%s at (%s)' % ('x'.join([str(x) for x in self.size]), ','.join([str(x) for x in self.lower]))
        except TypeError:
            return '%s at %s' % (self.size, self.lower)

    @staticmethod
    def to_box(value, resolution_hint=None):
        if value is None:
            result = AABox(0, resolution_hint)
        elif isinstance(value, AABox):
            result = value
        elif isinstance(value, int):
            if resolution_hint is None:
                result = AABox(0, value)
            else:
                size = [value] * (1 if math.ndims(resolution_hint) == 0 else len(resolution_hint))
                result = AABox(0, size)
        else:
            result = AABox(box)
        if resolution_hint is not None:
            assert_same_rank(len(resolution_hint), result, 'AABox rank does not match resolution.')
        return result


class AABoxGenerator(object):

    def __getitem__(self, item):
        if not isinstance(item, (tuple, list)):
            item = [item]
        lower = []
        upper = []
        for dim in item:
            assert isinstance(dim, slice)
            assert dim.step is None or dim.step == 1, "Box: step must be 1 but is %s" % dim.step
            lower.append(dim.start if dim.start is not None else -np.inf)
            upper.append(dim.stop if dim.stop is not None else np.inf)
        return AABox(lower, upper)


box = AABoxGenerator()


@struct.definition()
class Sphere(Geometry):

    def __init__(self, center, radius, **kwargs):
        Geometry.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def radius(self, radius):
        return math.as_tensor(radius)

    @struct.constant()
    def center(self, center):
        return math.as_tensor(center)

    def value_at(self, location):
        center = math.batch_align(self.center, 1, location)
        radius = math.batch_align(self.radius, 0, location)
        distance_squared = math.sum((location - center)**2, axis=-1, keepdims=True)
        bool_inside = distance_squared <= radius**2
        return math.to_float(bool_inside)

    @property
    def rank(self):
        return len(self.center)


@struct.definition()
class _Union(Geometry):

    def __init__(self, geometries, **kwargs):
        Geometry.__init__(self, **struct.kwargs(locals()))

    def __validate_geometries__(self):
        assert len(self.geometries) > 0
        rank = self.geometries[0].rank
        for g in self.geometries[1:]:
            assert g.rank == rank or g.rank is None or rank is None
        self.geometries = tuple(self.geometries)

    def value_at(self, points, collapse_dimensions=True):
        if len(self.geometries) == 1:
            result = self.geometries[0].value_at(points)
        else:
            result = math.max([geometry.value_at(points) for geometry in self.geometries], axis=0)
        return result

    @property
    def rank(self):
        if len(self.geometries) == 0:
            return None
        else:
            return self.geometries[0].rank

    @struct.constant()
    def geometries(self, geometries):
        return tuple(geometries)


def union(geometries):
    if len(geometries) == 0:
        return NO_GEOMETRY
    else:
        return _Union(geometries)


@struct.definition()
class _NoGeometry(Geometry):

    def rank(self):
        return None

    def value_at(self, location):
        return 0


NO_GEOMETRY = _NoGeometry()


def assert_same_rank(rank1, rank2, error_message):
    rank1_, rank2_ = _rank(rank1), _rank(rank2)
    if rank1_ is not None and rank2_ is not None:
        assert rank1_ == rank2_, 'Ranks do not match: %s and %s. %s' % (rank1_, rank2_, error_message)


def _rank(rank):
    if rank is None:
        return None
    if isinstance(rank, int):
        return rank
    if isinstance(rank, Geometry):
        return rank.rank
    else:
        return math.spatial_rank(rank)



import warnings

import numpy

import plotly.figure_factory as plotly_figures

from phi import math
from phi.physics.field import CenteredGrid, StaggeredGrid
from phi.viz.plot import FRONT, RIGHT, TOP

EMPTY_FIGURE = {'data': [{'z': None, 'type': 'heatmap'}]}


def dash_graph_plot(data, settings):
    # type: (object, dict) -> dict
    if data is None:
        return EMPTY_FIGURE

    if isinstance(data, numpy.ndarray):
        data = CenteredGrid(data)

    if isinstance(data, (CenteredGrid, StaggeredGrid)):
        component = settings.get('component', 'x')
        if data.rank == 1:
            return plot(data, settings)
        if data.rank == 2:
            if component == 'vec2' and data.component_count >= 2:
                return vector_field(data, settings)
            else:
                return heatmap(data, settings)
        if data.rank == 3:
            if component == 'vec2' and data.component_count >= 2:
                return vector_field(slice_2d(data, settings), settings)
            else:
                return heatmap(slice_2d(data, settings), settings)

    warnings.warn('No figure recipe for data %s' % data)
    return EMPTY_FIGURE


def heatmap(data, settings):
    assert isinstance(data, (StaggeredGrid, CenteredGrid))
    assert data.rank == 2
    batch = settings.get('batch', 0)
    component = settings.get('component', 'x')

    if isinstance(data, StaggeredGrid):
        if component == 'x':
            data = data.unstack()[-1]
        elif component == 'y':
            data = data.unstack()[-2]
        elif component == 'z':
            return EMPTY_FIGURE
        elif component == 'length':
            data = data.at_centers()
        else:
            raise ValueError(component)
    z = data.data[batch,...]
    z = reduce_component(z, component)
    y = data.points.data[0, :, 0, 0]
    x = data.points.data[0, 0, :, 1]
    return {'data': [{'x': x, 'y': y, 'z': z, 'type': 'heatmap'}]}


def slice_2d(field3d, settings):
    if isinstance(field3d, numpy.ndarray):
        field3d = CenteredGrid(field3d)
    if isinstance(field3d, StaggeredGrid):
        component = settings.get('component', 'length')
        if component in ('z', 'y', 'x'):
            field3d = field3d.unstack()[('z', 'y', 'x').index(component)]
        else:
            field3d = field3d.at_centers()
    assert isinstance(field3d, CenteredGrid) and field3d.rank == 3
    depth = settings.get('depth', 0)
    projection = settings.get('projection', FRONT)

    if projection == FRONT:
        # Remove Y axis
        data = field3d.data[:, :, min(depth, field3d.resolution[1]), :, :]
        field2d = CenteredGrid(data, box=field3d.box.without_axis(1))
    elif projection == RIGHT:
        # Remove X axis
        data = field3d.data[:, :, min(depth, field3d.resolution[2]), :, :]
        data = numpy.transpose(data, axes=(0, 2, 1, 3))
        field2d = CenteredGrid(data, box=field3d.box.without_axis(2))
    elif projection == TOP:
        # Remove Z axis
        data = field3d.data[:, min(depth, field3d.resolution[0]), :, :, :]
        field2d = CenteredGrid(data, box=field3d.box.without_axis(0))
    else:
        raise ValueError('Unknown projection: %s' % projection)
    return field2d


def plot(field1d, settings):
    assert isinstance(field1d, (CenteredGrid, StaggeredGrid))
    batch = settings.get('batch', 0)
    component = settings.get('component', 'x')
    if isinstance(field1d, StaggeredGrid):
        field1d = field1d.unstack()[0]
    assert isinstance(field1d, CenteredGrid)
    x = field1d.points.data[0, :, 0]
    data = field1d.data[min(field1d.resolution[0], batch), :, :]
    data = reduce_component(data, component)
    return {'data': [{'mode': 'markers+lines', 'type': 'scatter', 'x': x, 'y': data}]}


def reduce_component(tensor, component):
    clen = tensor.shape[-1]
    if clen == 1:
        return tensor[...,0]
    if component == 'x':
        return tensor[...,-1]
    if component == 'y':
        return tensor[...,-2]
    if component == 'z':
        if clen >= 3:
            return tensor[...,-3]
        else:
            return numpy.zeros_like(tensor[...,0])
    if component == 'length':
        return numpy.sqrt(numpy.sum(tensor ** 2, axis=-1, keepdims=False))
    if component == 'vec2':
        return tensor[...,-2:]


def vector_field(field2d, settings, draw_arrows_backward=True, max_resolution=40, max_lines=300, full_arrows=False):
    assert isinstance(field2d, (CenteredGrid, StaggeredGrid))
    if isinstance(field2d, StaggeredGrid):
        field2d = field2d.at_centers()
    assert isinstance(field2d, CenteredGrid)
    assert field2d.rank == 2

    batch = settings.get('batch', 0)
    batch = min(batch, field2d.data.shape[0])

    y, x = math.unstack(field2d.points.data[0,...,-2:], axis=-1)
    data_y, data_x = math.unstack(field2d.data[batch,...], -1)[-2:]

    while numpy.prod(x.shape) > max_resolution ** 2:
        y = y[::2, ::2]
        x = x[::2, ::2]
        data_y = data_y[::2, ::2]
        data_x = data_x[::2, ::2]

    y = y.flatten()
    x = x.flatten()
    data_y = data_y.flatten()
    data_x = data_x.flatten()

    if max_lines is not None and len(x) > max_lines:
        length = numpy.sqrt(data_y**2 + data_x**2)
        keep_indices = numpy.argsort(length)[-max_lines:]
        # size = numpy.max(field2d.box.size)
        # threshold = size * negligeable_threshold
        # keep_condition = (numpy.abs(data_x) > threshold) | (numpy.abs(data_y) > threshold)
        # keep_indices = numpy.where(keep_condition)
        y = y[keep_indices]
        x = x[keep_indices]
        data_y = data_y[keep_indices]
        data_x = data_x[keep_indices]

    if draw_arrows_backward:
        x -= data_x
        y -= data_y

    if full_arrows:
        result = plotly_figures.create_quiver(x, y, data_x, data_y, scale=1.0)  # 7 points per arrow
        result.update_xaxes(range=[field2d.box.get_lower(1), field2d.box.get_upper(1)])
        result.update_yaxes(range=[field2d.box.get_lower(0), field2d.box.get_upper(0)])
        return result
    else:
        lines_y = numpy.stack([y, y+data_y, [None]*len(x)], -1).flatten()  # 3 points per arrow
        lines_x = numpy.stack([x, x+data_x, [None]*len(x)], -1).flatten()
        return {
            'data': [
                {
                    'mode': 'lines',
                    'x': lines_x,
                    'y': lines_y,
                    'type': 'scatter',
                }
            ],
            'layout': {
                'xaxis': {'range': [field2d.box.get_lower(1), field2d.box.get_upper(1)]},
                'yaxis': {'range': [field2d.box.get_lower(0), field2d.box.get_upper(0)]},
            }
        }



import typing  # pylint: disable-msg = unused-import  # this is used in # type
from typing import Dict
import six


STRUCT_CLASSES = None  # type: tuple


_STRUCT_REGISTER = {}  # type: Dict[typing.Type, StructType]
_UNUSED_ITEMS = {}  # type: Dict[str, Item] # only temporary, before class decorator called


def definition():
    """
Required decorator for custom struct classes.
    """
    def decorator(cls):
        structtype = _build_type(cls)
        cls.__struct__ = structtype
        return cls
    return decorator


def variable(default=None, dependencies=(), holds_data=True):
    """
Required decorator for data_dict of custom structs.
The enclosing class must be decorated with struct.definition().
    :param default: default value passed to validation if no other value is specified
    :param dependencies: other items (string or reference for inherited constants_dict)
    :param holds_data: determines whether the variable is considered by data-related functions
    :return: read-only property
    """
    def decorator(validate):
        item = Item(validate.__name__, validate, True, default, dependencies, holds_data)
        _register_item(validate, item)
        return item
    return decorator


def constant(default=None, dependencies=(), holds_data=False):
    """
Required decorator for constants_dict of custom structs.
The enclosing class must be decorated with struct.definition().
    :param default: default value passed to validation if no other value is specified
    :param dependencies: other items (string or reference for inherited constants_dict)
    :param holds_data: determines whether the constant is considered by data-related functions
    :return: read-only property
    """
    def decorator(validate):
        item = Item(validate.__name__, validate, False, default, dependencies, holds_data)
        _register_item(validate, item)
        return item
    return decorator


def derived():
    """
Derived properties work similar to @property but can be easily broadcast across many instances.
    :return: read-only property
    """
    def decorator(getter):
        return DerivedProperty(getter.__name__, getter)
    return decorator


def _register_item(_function, item):
    _UNUSED_ITEMS[item.name] = item


def _build_type(cls):
    assert cls not in _STRUCT_REGISTER
    items = {}
    for attribute in dir(cls):
        if attribute in _UNUSED_ITEMS:
            items[attribute] = _UNUSED_ITEMS.pop(attribute)
    for base in cls.__bases__:
        if base not in STRUCT_CLASSES and base in _STRUCT_REGISTER:
            basetype = _STRUCT_REGISTER[base]
            for item in basetype.items:
                if item.name not in items:
                    items[item.name] = item
    structtype = StructType(cls, items)
    _STRUCT_REGISTER[cls] = structtype
    return structtype


def get_type(struct_class):
    """
    :rtype: StructType
    """
    return _STRUCT_REGISTER[struct_class]


class StructType(object):
    """
One StructType is associated with each defined struct (subclass of Struct) and stored in the _STRUCT_REGISTER.
    """

    def __init__(self, struct_class, item_dict):
        self.struct_class = struct_class
        self.item_dict = item_dict
        self.items = _order_by_dependencies(item_dict, self)
        self.variables = tuple(filter(lambda item: item.is_variable, self.items))
        self.constants = tuple(filter(lambda item: not item.is_variable, self.items))

    def find(self, name):
        return self.item_dict[name]

    def validate(self, struct):
        for item in self.items:
            item.validate(struct)

    @property
    def item_names(self):
        return [item.name for item in self.items]

    def __repr__(self):
        return self.struct_class.__name__


class Item(object):
    """
Represents an item type of a struct, a variable or a constant.
    """

    def __init__(self, name, validation_function, is_variable, default_value, dependencies, holds_data):
        assert callable(validation_function) or validation_function is None
        self.name = name
        self.validation_function = validation_function
        self.is_variable = is_variable
        self.default_value = default_value
        self.dependencies = dependencies
        self.holds_data = holds_data
        self.owner = None

    def set(self, struct, value):
        try:
            setattr(struct, '_' + self.name, value)
        except AttributeError:
            raise AttributeError("can't modify struct %s because item %s cannot be set." % (struct, self))

    def get(self, struct):
        return getattr(struct, '_' + self.name)

    def validate(self, struct):
        if self.validation_function is not None:
            old_val = self.get(struct)
            new_val = self.validation_function(struct, old_val)
            self.set(struct, new_val)

    def __get__(self, instance, owner):
        if instance is not None:
            return getattr(instance, '_'+self.name)
        else:
            self.owner = owner
            return self

    def __call__(self, obj):
        assert self.owner is not None
        from .functions import map
        return map(lambda x: getattr(x, '_'+self.name), obj, leaf_condition=lambda x: isinstance(x, self.owner))

    def __set__(self, instance, value):
        raise AttributeError('Struct variables and constants are read-only.')

    def __delete__(self, instance):
        raise AttributeError('Struct variables and constants are read-only.')

    def __repr__(self):
        return self.name


def CONSTANTS(item): return not item.is_variable
def VARIABLES(item): return item.is_variable
def DATA(item): return item.holds_data
ALL_ITEMS = None


class DerivedProperty(object):

    def __init__(self, name, getter):
        self.name = name
        self.getter = getter

    def __get__(self, instance, owner):
        if instance is not None:
            return self.getter(instance)
        else:
            self.owner = owner
            return self

    def __call__(self, obj):
        assert self.owner is not None
        from .functions import map
        return map(lambda x: self.getter(x), obj, leaf_condition=lambda x: isinstance(x, self.owner))

    def __set__(self, instance, value):
        raise AttributeError('Derived constants_dict cannot be set.')

    def __delete__(self, instance):
        raise AttributeError('Derived constants_dict cannot be deleted.')

    def __repr__(self):
        return self.name


def _order_by_dependencies(item_dict, owner):
    result = []
    for item in item_dict.values():
        _recursive_deps_add(item, item_dict, result, owner)
    return result


def _recursive_deps_add(item, item_dict, result_list, owner):
    if item in result_list: return
    dependencies = _get_dependencies(item, item_dict, owner)
    for dependency in dependencies:
        _recursive_deps_add(dependency, item_dict, result_list, owner)
    result_list.append(item)


def _get_dependencies(item, item_dict, owner):
    dependencies = _resolve_dependencies(item.dependencies, item_dict, owner)
    unique_dependencies = set(dependencies)
    assert len(unique_dependencies) == len(dependencies), 'Duplicate dependencies in item %s' % item
    return unique_dependencies


def _resolve_dependencies(dependency, item_dict, owner):
    if dependency is None: return []
    if isinstance(dependency, six.string_types):
        try:
            return [item_dict[dependency]]
        except KeyError:
            raise DependencyError('Declared dependency "%s" does not exist on struct %s. Properties: %s' % (dependency, owner, tuple(item_dict.keys())))
    if isinstance(dependency, Item): return [item_dict[dependency.name]]
    if isinstance(dependency, (tuple, list)):
        return sum([_resolve_dependencies(dep, item_dict, owner) for dep in dependency], [])
    raise ValueError('Cannot resolve dependency "%s". Available items: %s' % (dependency, item_dict.keys()))


class DependencyError(Exception):

    def __init__(self, *args):
        Exception.__init__(self, *args)



import six

from .structdef import DATA, ALL_ITEMS
from .struct import isstruct, copy_with, equal, to_dict
from .context import unsafe


def flatten(struct, leaf_condition=None, trace=False, item_condition=DATA):
    result = []
    def map_leaf(value):
        result.append(value)
        return value
    with unsafe(): map(map_leaf, struct, leaf_condition, recursive=True, trace=trace, item_condition=item_condition)
    return result


def names(struct, leaf_condition=None, full_path=True, basename=None, separator='.'):
    def to_name(trace):
        if not full_path:
            return trace.name if basename is None else basename + separator + trace.name
        else:
            return trace.path(separator) if basename is None else basename + separator + trace.path(separator)
    with unsafe():
        return map(to_name, struct, leaf_condition, recursive=True, trace=True)


def zip(structs, leaf_condition=None, item_condition=DATA, zip_parents_if_incompatible=False):
    # pylint: disable-msg = redefined-builtin
    assert len(structs) > 0
    first = structs[0]
    if isstruct(first, leaf_condition):
        for struct in structs[1:]:
            if set(to_dict(struct, item_condition=item_condition).keys()) != set(to_dict(first, item_condition=item_condition).keys()):
                if zip_parents_if_incompatible:
                    return LeafZip(structs)
                else:
                    raise IncompatibleStructs('Cannot zip %s and %s because keys vary:\n%s\n%s' % (struct, first, to_dict(struct, item_condition=item_condition).keys(), to_dict(first, item_condition=item_condition).keys()))

    if not isstruct(first, leaf_condition):
        return LeafZip(structs)

    dicts = [to_dict(struct, item_condition=item_condition) for struct in structs]
    keys = dicts[0].keys()
    new_dict = {}
    for key in keys:
        values = [d[key] for d in dicts]
        values = zip(values, leaf_condition, item_condition=item_condition, zip_parents_if_incompatible=zip_parents_if_incompatible)
        new_dict[key] = values
    with unsafe():
        return copy_with(first, new_dict)


class LeafZip(object):
    """
Created by struct.zip to replace data.
    """

    def __init__(self, values):
        self.values = values

    def __getitem__(self, item):
        return self.values[item]

    def __repr__(self):
        return repr(self.values)

    def __str__(self):
        return str(self.values)


class IncompatibleStructs(Exception):
    """
Thrown when two or more structs are required to have the same structure but do not.
    """
    def __init__(self, *args):
        Exception.__init__(self, *args)


def map(function, struct, leaf_condition=None, recursive=True, trace=False, item_condition=DATA):
    # pylint: disable-msg = redefined-builtin
    if trace is True:
        trace = Trace(struct, None, None)
    if not isstruct(struct, leaf_condition):
        if trace is False:
            if isinstance(struct, LeafZip):
                return function(*struct.values)
            else:
                return function(struct)
        else:
            return function(trace)
    else:
        old_values = to_dict(struct, item_condition=item_condition)
        new_values = {}
        if not recursive:
            leaf_condition = lambda x: True
        for key, value in old_values.items():
            new_values[key] = map(function, value, leaf_condition, recursive,
                                  Trace(value, key, trace) if trace is not False else False,
                                  item_condition=item_condition)

        return copy_with(struct, new_values)


class Trace(object):
    """
Used in struct.map if trace=True.
Trace objects can be used to reference a specific item of a struct or sub-struct as well as gather information about it.
    """

    def __init__(self, value, key, parent_trace):
        self.value = value
        self.key = key
        self.parent = parent_trace

    @property
    def name(self):
        if self.key is None:
            return None
        if isinstance(self.key, six.string_types):
            return self.key
        else:
            return str(self.key)

    def path(self, separator='.'):
        if self.parent is not None and self.parent.key is not None:
            return self.parent.path(separator) + separator + self.name
        else:
            return self.name

    def __repr__(self):
        return "%s = %s" % (self.path(), self.value)

    def find_in(self, base_struct):
        if self.parent is not None and self.parent.key is not None:
            base_struct = self.parent.find_in(base_struct)
        attrs = to_dict(base_struct, item_condition=ALL_ITEMS)
        return attrs[self.key]


def compare(structs, leaf_condition=None, recursive=True, item_condition=ALL_ITEMS):
    if len(structs) <= 1: return []
    result = set()
    def check(trace):
        value = trace.value
        for other in structs[1:]:
            try:
                other_value = trace.find_in(other)
                if not equal(value, other_value):
                    result.add(trace)
            except (ValueError, KeyError, TypeError):
                result.add(trace)
    with unsafe(): map(check, structs[0], leaf_condition=leaf_condition, recursive=recursive, trace=True, item_condition=item_condition)
    return result


def print_differences(struct1, struct2):
    if not isstruct(struct1) or not isstruct(struct2):
        if not equal(struct1, struct2):
            print('Values not equal: "%s" and "%s".' % (struct1, struct2))
        return
    items1 = to_dict(struct1)
    items2 = to_dict(struct2)
    tested_keys = []
    for key1 in items1.keys():
        if key1 not in items2:
            print('Item "%s" is missing from %s.' % (key1, struct2))
        else:
            if not equal(items1[key1], items2[key1]):
                print('Item "%s" differs between %s and %s.' % (key1, struct1, struct2))
            print_differences(items1[key1], items2[key1])
        tested_keys.append(key1)
    for key2 in items2.keys():
        if key2 not in tested_keys:
            print('Item "%s" is missing from %s.' % (key2, struct1))


def mappable(leaf_condition=None, recursive=True, item_condition=DATA, unsafe_context=False):
    def decorator(function):
        def broadcast_function(obj, *args, **kwargs):
            def function_with_args(x): return function(x, *args, **kwargs)
            if unsafe_context:
                with unsafe():
                    result = map(function_with_args, obj, leaf_condition=leaf_condition, recursive=recursive, item_condition=item_condition)
            else:
                result = map(function_with_args, obj, leaf_condition=leaf_condition, recursive=recursive, item_condition=item_condition)
            return result
        return broadcast_function
    return decorator



from numpy import log10


class EditableValue(object):

    def __init__(self, name, type, initial_value, category, minmax, is_linear):
        self.name = name
        self.type = type
        self.initial_value = initial_value
        self.category = category
        self.minmax = minmax
        self.is_linear = is_linear

    @property
    def min_value(self):
        return self.minmax[0]

    @property
    def max_value(self):
        return self.minmax[1]


class EditableFloat(EditableValue):

    def __init__(self, name, initial_value, minmax=None, category=None, log_scale=None):
        if minmax is not None:
            assert len(minmax) == 2, 'minmax must be pair (min, max)'

        if log_scale is None:
            if minmax is None:
                log_scale = True
            else:
                log_scale = minmax[1] / float(minmax[0]) > 10

        if not minmax:
            if log_scale:
                magn = log10(initial_value)
                minmax = (10.0**(magn-3.2), 10.0**(magn+2.2))
            else:
                if initial_value == 0.0:
                    minmax = (-10.0, 10.0)
                elif initial_value > 0:
                    minmax = (0., 4. * initial_value)
                else:
                    minmax = (2. * initial_value, -2. * initial_value)
        else:
            minmax = (float(minmax[0]), float(minmax[1]))
        EditableValue.__init__(self, name, 'float', initial_value, category, minmax, not log_scale)

    @property
    def use_log_scale(self):
        return not self.is_linear


class EditableInt(EditableValue):

    def __init__(self, name, initial_value, minmax=None, category=None):
        if not minmax:
            if initial_value == 0:
                minmax = (-10, 10)
            elif initial_value > 0:
                minmax = (0, 4*initial_value)
            else:
                minmax = (2 * initial_value, -2 * initial_value)
        EditableValue.__init__(self, name, 'int', initial_value, category, minmax, True)


class EditableBool(EditableValue):

    def __init__(self, name, initial_value, category=None):
        EditableValue.__init__(self, name, 'bool', initial_value, category, (False, True), True)


class EditableString(EditableValue):

    def __init__(self, name, initial_value, category=None, rows=20):
        EditableValue.__init__(self, name, 'text', initial_value, category, ('', 'A'*rows), True)

    @property
    def rows(self):
        return len(self.max_value)



from phi import struct
from phi.geom.geometry import Geometry
from .field.effect import FieldEffect, GeometryMask
from .physics import State, Physics
from .material import Material, CLOSED


@struct.definition()
class Obstacle(State):

    def __init__(self, geometry, material=CLOSED, velocity=0, tags=('obstacle',), **kwargs):
        State.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def geometry(self, geometry):
        assert isinstance(geometry, Geometry)
        return geometry

    @struct.constant(default=CLOSED)
    def material(self, material):
        assert isinstance(material, Material)
        return material

    @struct.constant(default=0)
    def velocity(self, velocity): return velocity


class GeometryMovement(Physics):

    def __init__(self, geometry_function):
        Physics.__init__(self)
        self.geometry_at = geometry_function

    def step(self, obj, dt=1.0, **dependent_states):
        next_geometry = self.geometry_at(obj.age + dt)
        h = 1e-2 * dt if dt > 0 else 1e-2
        perturbed_geometry = self.geometry_at(obj.age + dt + h)
        velocity = (perturbed_geometry.center - next_geometry.center) / h
        if isinstance(obj, Obstacle):
            return obj.copied_with(geometry=next_geometry, velocity=velocity, age=obj.age + dt)
        if isinstance(obj, FieldEffect):
            field = obj.field
            assert isinstance(field, GeometryMask)
            assert len(field.geometries) == 1
            return obj.copied_with(field=obj.field.copied_with(geometries=(next_geometry,)), age=obj.age + dt)
